#include <pic.h>	// 12C508

#define Version 0012

/*
   SX programming HW driver V1.2 (for COMPIC-1 HW with V2.03+ SW)
   [*] Compile with optimization ON

   Copyright (C)1999 Nick Walter <nick@msh.lv>
*/

__CONFIG(WDTE|UNPROTECT|INTRC);
__IDLOC(Version);

#define uint unsigned int
#define	PORTBIT(adr, bit) ((unsigned)(&adr)*8+(bit))
#define byte(adr)	  (*((char*)&adr))
#define word(adr)	  (*((uint*)&adr))

static bit DAT @ PORTBIT(GPIO,0);
static bit CLK @ PORTBIT(GPIO,1);
static bit SXDAT @ PORTBIT(GPIO,2);
static bit SXOSC1 @ PORTBIT(GPIO,4);

#define _SXDATO   0b101011
#define __SXDATO    101011B
#define _SXDATI   0b101111
#define _DATO     0b101110
#define _DATI     0b101111
#define _option 0b10011011

bit turbo;
bit sxread;
bit cturbo;
uint sxreg;
uint sxbuf;
char sxloop;
char dly;
char si;
char rxbuf;
char progloop;
char result;
char wdly,rdly;

char progbuf[8];

#define FAILED  0x5A
#define OK      0xA5
#define CPrefix 0xA0
enum {cmECHO,cmVER,cmENTERISP1,cmENTERISP2,cmCHECKSXSYNC,cmEXITISP,cmSXIO,
      cmBUFSIZE,cmGETBUF,cmPUTBUF,cmREADSX,cmPROGSX,cmSXBUF,cmRESULT,
      cmPROGTIME};

#define CTIME 32
#define CSYNC (CTIME+(CTIME>>1))

#define true  1
#define false 0

#define delay(x) {dly=x/3; while(--dly);}
#define clrwdt() asm("clrwdt")
#define rlf(f) asm("rlf _"___mkstr(f))
#define nop() asm("nop")
#define wait(b) do clrwdt(); while(!(b))
#define copybit(src,dst) {if(src) dst=1; if(!src) dst=0;}
#define copyint(src,dst) (byte(dst)=byte(src),byte(dst+1)=byte(src+1))

#define SXWrite() {asm("movlw "___mkstr(__SXDATO)); \
      		   asm("btfss _sxreg+1,7"); \
      		   asm("tris 6"); \
		   dly=wdly; while(--dly); \
	           TRIS=_SXDATI; \
		   sxreg<<=1;}

#define SXRead()  {dly=rdly; while(--dly); \
		   CARRY=SXDAT; \
		   rlf(sxreg); rlf (sxreg+1); \
	           while(!SXDAT);}

void fastcall CheckSXSync(void)
{char i;
	// wait sync
	i=3;
	do {
	    do {TMR0=0; while(SXDAT);} while(TMR0<CSYNC);
	    while(!SXDAT);
	    // get cycle time for timing calibration
	    TMR0=0; while(SXDAT); byte(sxbuf)=TMR0;
	    while(!SXDAT);
	    TMR0=0; while(SXDAT); byte(sxbuf+1)=TMR0;
	    while(!SXDAT);
	    clrwdt();
	   } while(--i);
	result=OK;
}

void fastcall EnterISP1(void)
{
	TRIS=_SXDATO;
	SXOSC1=1; SXDAT=1; delay(24);
	SXDAT=0;
	si=10;
	do {SXOSC1=0; delay(15);
	    SXOSC1=1; delay(15);} while(--si);
	SXDAT=1; delay(24);
	SXOSC1=0;
	TRIS=_SXDATI;
	result=OK;
}

void fastcall EnterISP2(void)
{
	TRIS=_SXDATO;
	SXOSC1=0; SXDAT=0;
	si=0; do {delay(9);} while(--si);
	SXDAT=1; delay(24);
	TRIS=_SXDATI;
	result=OK;
}

void fastcall ExitISP(void)
{
	SXOSC1=1; delay(24); SXOSC1=0;
	result=OK;
}

void fastcall sxcmd(void)
{char i;
	SXDAT=0; i=0; if(!turbo) cturbo=false;
do {
	copyint(sxbuf,sxreg);
	// wait sync
	// repeat command without resync to speed up programming time
	i|=4;
	do {TMR0=CSYNC; if(!cturbo) TMR0=0;
	    while(SXDAT);} while(TMR0<CSYNC);
	// write command to sx
	do {SXWrite(); while(SXDAT);} while(--i); i=11;
	if(sxread) // read data from sx
	 {byte(sxreg)=0;
	  do {SXRead(); while(SXDAT);} while(--i);
	  SXRead();
	  copyint(sxreg,sxbuf);
	 }
	else // write data to sx
	 {do {SXWrite(); while(SXDAT);} while(--i);
	  SXWrite();
	 }
	clrwdt();
	cturbo=true;
   } while(--sxloop);
}

void fastcall waitstart(void)
{
WaitStart:
	do {wait(CLK); wait(!CLK);} while(!DAT);
	do {if(CLK) goto WaitStart; clrwdt();} while(DAT);
	wait(CLK);
}

void fastcall RxByte(void)
{
	si=8;
	do {wait(!CLK); rxbuf=(rxbuf<<1)|DAT;
	    wait(CLK);} while(--si);
}

void fastcall TxByte(char txbuf)
{
	TRIS=_DATO; si=8;
	do {txbuf<<=1;
	    copybit(CARRY,DAT);
	    wait(!CLK); wait(CLK);} while(--si);
	TRIS=_DATI;
}

void main(void)
{char n;
 char *pbuf;
	OSCCAL=0xF0;
	TRIS=_DATI;
	GPIO=_DATI;
	OPTION=_option;
	result=FAILED;
	turbo=false; wdly=6; rdly=4;
MainLoop:
	waitstart();
	RxByte();
	rxbuf-=CPrefix;
	switch(rxbuf)
	{case cmECHO:
	  RxByte();
	  TxByte(rxbuf);
	  break;
	 case cmVER:
	  TxByte(0xVersion);
	  break;
	 case cmENTERISP1:
	  EnterISP1(); // set's result
	  break;
	 case cmENTERISP2:
	  EnterISP2(); // set's result
	  break;
	 case cmCHECKSXSYNC:
	  CheckSXSync(); // set's result
	  break;
	 case cmEXITISP:
	  ExitISP(); // set's result
	  break;
	 case cmSXIO:
	  RxByte();
	  byte(sxloop)=rxbuf;
	  RxByte();
	  byte(sxbuf)=rxbuf;
	  RxByte();
	  byte(sxbuf+1)=rxbuf;
	  sxread=!sxloop;
	  if(sxread) sxloop++;
	  cturbo=false;
	  sxcmd();
	  break;
	 case cmBUFSIZE:
	  TxByte(sizeof(progbuf));
	  break;
	 case cmGETBUF:
	  RxByte();
	  pbuf=progbuf;
	  do TxByte(*pbuf++); while(--rxbuf);
	  break;
	 case cmPUTBUF:
	  RxByte();
	  n=rxbuf; pbuf=progbuf;
	  do {RxByte(); *pbuf++=rxbuf;} while(--n);
	  break;
	 case cmREADSX:
	  RxByte();
	  n=rxbuf; pbuf=progbuf;
	  cturbo=false;
	  do {sxloop++;
	      sxbuf=0x6FFF; // READ
	      sxread=true;
	      sxcmd();
	      word(*pbuf++)=sxbuf; pbuf++;
	      sxloop++;
	      sxbuf=0x7FFF; // INC
	      sxread=false;
	      sxcmd();
	     } while(--n);
	  result=OK;
	  break;
	 case cmPROGSX:
	  RxByte(); n=rxbuf;
	  RxByte(); progloop=rxbuf;
	  pbuf=progbuf;
	  cturbo=false;
	  result=OK;
	  do {sxloop++;
	      sxbuf=word(*pbuf);
	      byte(sxbuf+1)|=0x40; // LOAD
	      sxread=false;
	      sxcmd();
	      sxloop=progloop;
	      sxbuf=0x5FFF;	   // PROG
	      sxcmd();
	      sxloop++;
	      sxbuf=0x6FFF;	   // READ
	      sxread=true;
	      sxcmd();
	      if(sxbuf!=word(*pbuf++)) {result=FAILED; break;} // verify failed
	      pbuf++;
	      sxloop++;
	      sxbuf=0x7FFF;	   // INC
	      sxread=false;
	      cturbo=false;
	      sxcmd();
	     } while(--n);
	  break;
	 case cmSXBUF:
	  TxByte(result);
	  TxByte(byte(sxbuf));
	  TxByte(byte(sxbuf+1));
	  break;
 	 case cmRESULT:
	  TxByte(result);
	  break;
	 case cmPROGTIME:
	  RxByte();
	  turbo=rxbuf;
	  rxbuf&=0xFC;
	  if(rxbuf) OSCCAL=rxbuf;
	  RxByte();
	  if(rxbuf) wdly=rxbuf;
	  RxByte();
	  if(rxbuf) rdly=rxbuf;
	}
	goto MainLoop;
}
